import random
import string

import ipywidgets as widgets
from IPython.display import display
from traitlets import List, Unicode

from d3vis_ipynb.base_widget import BaseWidget


@widgets.register
class MatrixLayout(BaseWidget):
    _view_name = Unicode("MatrixLayoutView").tag(sync=True)
    _model_name = Unicode("MatrixLayoutModel").tag(sync=True)

    matrix = List().tag(sync=True)
    grid_areas = List().tag(sync=True)
    grid_template_areas = Unicode().tag(sync=True)
    style = Unicode().tag(sync=True)

    def __init__(self, matrix, **kwargs):
        self._is_displayed = False
        self._widgets_to_display = {}
        self._check_matrix_format(matrix)
        self.matrix = matrix

        self._all_widgets = []
        self.positions_hashs = {}
        self.grid_areas = []
        for num in self.all_numbers:
            random_string = "".join(
                random.choice(string.ascii_letters) for i in range(10)
            )
            self.positions_hashs[num] = random_string
            self.grid_areas.append(random_string)

        self.grid_template_areas = ""
        for row in matrix:
            self.grid_template_areas = self.grid_template_areas + '\n"'
            for num in row:
                self.grid_template_areas = (
                    self.grid_template_areas + self.positions_hashs[num] + " "
                )
            self.grid_template_areas = self.grid_template_areas + '"'
        super().__init__(**kwargs)

    def _check_matrix_format(self, matrix):
        def not_list_of_lists():
            raise Exception("Matrix format must be a list of lists of integers")

        if any([type(row) is not list for row in matrix]):
            not_list_of_lists()
        self.all_numbers = []
        for row in matrix:
            for item in row:
                if type(item) is not int:
                    not_list_of_lists()
                if item < 0:
                    raise Exception("All integers must be positives")
                if not item in self.all_numbers:
                    self.all_numbers.append(item)

        first_row_len = len(matrix[0])
        if any([len(row) != first_row_len for row in matrix]):
            raise Exception("All rows must have the same size")

        self.all_numbers.sort()
        for i in range(1, len(self.all_numbers)):
            if self.all_numbers[i] - self.all_numbers[i - 1] != 1:
                raise Exception("All numbers must be in sequence.")

        self._check_if_has_only_rects(matrix)

    def _check_if_has_only_rects(self, matrix):
        def not_rects():
            raise Exception("Matrix must contain only unduplicate rectangles.")

        all_positions = {}

        for i in range(len(matrix)):
            row = matrix[i]
            for j in range(len(row)):
                item = row[j]
                position = (i, j)
                if not item in all_positions.keys():
                    all_positions[item] = []
                all_positions[item].append(position)

        for num in all_positions.keys():
            rows = {}
            num_positions = all_positions[num]
            num_positions.sort()
            for position in num_positions:
                if not position[0] in rows.keys():
                    rows[position[0]] = []
                rows[position[0]].append(position[1])
            # Check if there are numbers outside the rectangle
            first_row = list(rows.keys())[0]
            for i in range(1, len(rows[first_row])):
                if rows[first_row][i] - rows[first_row][i - 1] != 1:
                    not_rects()
            rows_keys = list(rows.keys())
            for i in range(1, len(rows_keys)):
                if rows_keys[i] - rows_keys[i - 1] != 1:
                    not_rects()
                if len(rows[rows_keys[i]]) != len(rows[first_row]):
                    not_rects()
                if rows[rows_keys[i]][0] != rows[first_row][0]:
                    not_rects()
                if rows[rows_keys[i]][-1] != rows[first_row][-1]:
                    not_rects()

    def add(self, widget, position: int):
        self._all_widgets.append(widget)
        if self._is_displayed:
            widget.elementId = self.positions_hashs[position]
            display(widget)
        else:
            self._widgets_to_display[position] = widget

    def _ipython_display_(self):
        self._is_displayed = True
        for key in self._widgets_to_display.keys():
            widget = self._widgets_to_display[key]
            widget.elementId = self.positions_hashs[key]
            display(widget)
        super()._ipython_display_()



import ipywidgets as widgets
import numpy as np
import pandas as pd
from traitlets import Float, List, Unicode

from d3vis_ipynb.base_widget import BaseWidget


@widgets.register
class BarPlot(BaseWidget):
    _view_name = Unicode("BarPlotView").tag(sync=True)
    _model_name = Unicode("BarPlotModel").tag(sync=True)

    dataRecords = List([]).tag(sync=True)
    x = Unicode().tag(sync=True)
    y = Unicode().tag(sync=True)
    hue = Unicode().tag(sync=True)

    def __init__(self, data, **kwargs):
        self.data = data
        super().__init__(**kwargs)

    @property
    def data(self):
        return pd.DataFrame.from_records(self.dataRecords)

    @data.setter
    def data(self, val):
        self.dataRecords = val.to_dict(orient="records")


@widgets.register
class DecisionPlot(BaseWidget):
    _view_name = Unicode("DecisionPlotView").tag(sync=True)
    _model_name = Unicode("DecisionPlotModel").tag(sync=True)

    dataRecords = List([]).tag(sync=True)
    baseValue = Float().tag(sync=True)

    def __init__(
        self,
        explanation,
        **kwargs,
    ):
        self.explanation = explanation
        super().__init__(**kwargs)

    @property
    def explanation(self):
        return self.explanation

    @explanation.setter
    def explanation(self, val):
        valuesArray = np.transpose(val.values).tolist()
        records = []
        for i in range(len(val.feature_names)):
            records.append(
                {"feature_names": val.feature_names[i], "values": valuesArray[i]}
            )
        self.baseValue = val.base_values[0]
        self.dataRecords = records


@widgets.register
class HistogramPlot(BaseWidget):
    _view_name = Unicode("HistogramPlotView").tag(sync=True)
    _model_name = Unicode("HistogramPlotModel").tag(sync=True)

    dataRecords = List([]).tag(sync=True)
    x = Unicode().tag(sync=True)

    def __init__(self, data, **kwargs):
        self.data = data
        super().__init__(**kwargs)

    @property
    def data(self):
        return pd.DataFrame.from_records(self.dataRecords)

    @data.setter
    def data(self, val):
        self.dataRecords = val.to_dict(orient="records")


@widgets.register
class LinearPlot(BaseWidget):
    _view_name = Unicode("LinearPlotView").tag(sync=True)
    _model_name = Unicode("LinearPlotModel").tag(sync=True)

    dataRecords = List([]).tag(sync=True)
    x = Unicode().tag(sync=True)
    y = Unicode().tag(sync=True)
    hue = Unicode().tag(sync=True)
    clickedValue = Unicode().tag(sync=True)
    selectedValuesRecords = List([]).tag(sync=True)

    def __init__(self, data, **kwargs):
        self.data = data
        self.selectedValues = pd.DataFrame()
        super().__init__(**kwargs)

    @property
    def data(self):
        return pd.DataFrame.from_records(self.dataRecords)

    @data.setter
    def data(self, val):
        self.dataRecords = val.to_dict(orient="records")

    @property
    def selectedValues(self):
        return pd.DataFrame.from_records(self.selectedValuesRecords)

    @selectedValues.setter
    def selectedValues(self, val):
        self.selectedValuesRecords = val.to_dict(orient="records")

    def on_select_values(self, callback):
        self.observe(callback, names=["selectedValuesRecords"])

    def on_click_value(self, callback):
        self.observe(callback, names=["clickedValue"])


@widgets.register
class RidgelinePlot(BaseWidget):
    _view_name = Unicode("RidgelinePlotView").tag(sync=True)
    _model_name = Unicode("RidgelinePlotModel").tag(sync=True)

    dataRecords = List([]).tag(sync=True)
    xAxes = List([]).tag(sync=True)

    def __init__(self, data, **kwargs):
        self.data = data
        super().__init__(**kwargs)

    @property
    def data(self):
        return pd.DataFrame.from_records(self.dataRecords)

    @data.setter
    def data(self, val):
        self.dataRecords = val.to_dict(orient="records")


@widgets.register
class ScatterPlot(BaseWidget):
    _view_name = Unicode("ScatterPlotView").tag(sync=True)
    _model_name = Unicode("ScatterPlotModel").tag(sync=True)

    dataRecords = List([]).tag(sync=True)
    x = Unicode().tag(sync=True)
    y = Unicode().tag(sync=True)
    hue = Unicode().tag(sync=True)
    clickedValue = Unicode().tag(sync=True)
    selectedValuesRecords = List([]).tag(sync=True)

    def __init__(self, data, **kwargs):
        self.data = data
        self.selectedValues = pd.DataFrame()
        super().__init__(**kwargs)

    @property
    def data(self):
        return pd.DataFrame.from_records(self.dataRecords)

    @data.setter
    def data(self, val):
        self.dataRecords = val.to_dict(orient="records")

    @property
    def selectedValues(self):
        return pd.DataFrame.from_records(self.selectedValuesRecords)

    @selectedValues.setter
    def selectedValues(self, val):
        self.selectedValuesRecords = val.to_dict(orient="records")

    def on_select_values(self, callback):
        self.observe(callback, names=["selectedValuesRecords"])

    def on_click_value(self, callback):
        self.observe(callback, names=["clickedValue"])


@widgets.register
class WaterfallPlot(BaseWidget):
    _view_name = Unicode("WaterfallPlotView").tag(sync=True)
    _model_name = Unicode("WaterfallPlotModel").tag(sync=True)

    dataRecords = List([]).tag(sync=True)
    x = Unicode().tag(sync=True)
    y = Unicode().tag(sync=True)
    baseValue = Float().tag(sync=True)

    def __init__(
        self,
        data=pd.DataFrame(),
        x="values",
        y="feature_names",
        explanation=None,
        baseValue=0,
        **kwargs,
    ):
        self.x = x
        self.y = y
        self.baseValue = baseValue
        if explanation != None and not data.empty:
            raise Exception("Initialize with explanation or data, not both.")
        elif not data.empty:
            self.data = data
        else:
            self.explanation = explanation
        super().__init__(**kwargs)

    @property
    def data(self):
        return pd.DataFrame.from_records(self.dataRecords)

    @data.setter
    def data(self, val):
        self.dataRecords = val.to_dict(orient="records")

    @property
    def explanation(self):
        return self.explanation

    @explanation.setter
    def explanation(self, val):
        df = pd.DataFrame()
        df.insert(0, self.x, val.values)
        df.insert(0, self.y, val.feature_names)
        self.baseValue = val.base_values
        self.data = df



import ipywidgets as widgets
import pandas as pd
from traitlets import Float, List, Unicode, Bool

from d3vis_ipynb.base_widget import BaseWidget


class TextBaseWidget(BaseWidget):
    value = Unicode().tag(sync=True)
    placeholder = Unicode().tag(sync=True)
    description = Unicode().tag(sync=True)
    disabled = Bool().tag(sync=True)


@widgets.register
class Button(BaseWidget):
    _view_name = Unicode("ButtonView").tag(sync=True)
    _model_name = Unicode("ButtonModel").tag(sync=True)

    description = Unicode().tag(sync=True)
    disabled = Bool().tag(sync=True)
    _clicked = Bool().tag(sync=True)

    def on_click(self, callback):
        self.observe(callback, names=["_clicked"])


@widgets.register
class Dropdown(BaseWidget):
    _view_name = Unicode("DropdownView").tag(sync=True)
    _model_name = Unicode("DropdownModel").tag(sync=True)

    dataRecords = List([]).tag(sync=True)
    variable = Unicode().tag(sync=True)
    description = Unicode().tag(sync=True)
    options = List().tag(sync=True)
    value = Unicode().tag(sync=True)
    disabled = Bool().tag(sync=True)
    _clicked = Bool().tag(sync=True)

    def __init__(self, data=pd.DataFrame(), **kwargs):
        self.data = data
        super().__init__(**kwargs)

    @property
    def data(self):
        return pd.DataFrame.from_records(self.dataRecords)

    @data.setter
    def data(self, val):
        self.dataRecords = val.to_dict(orient="records")

    def on_select(self, callback):
        self.observe(callback, names=["value"])


@widgets.register
class Input(TextBaseWidget):
    _view_name = Unicode("InputView").tag(sync=True)
    _model_name = Unicode("InputModel").tag(sync=True)

    def on_text_changed(self, callback):
        self.observe(callback, names=["value"])


@widgets.register
class RangeSlider(BaseWidget):
    _view_name = Unicode("RangeSliderView").tag(sync=True)
    _model_name = Unicode("RangeSliderModel").tag(sync=True)

    dataRecords = List([]).tag(sync=True)
    variable = Unicode().tag(sync=True)
    step = Float().tag(sync=True)
    description = Unicode().tag(sync=True)
    fromValue = Float().tag(sync=True)
    toValue = Float().tag(sync=True)
    minValue = Float().tag(sync=True)
    maxValue = Float().tag(sync=True)

    def __init__(self, data=pd.DataFrame(), **kwargs):
        self.data = data
        super().__init__(**kwargs)

    @property
    def data(self):
        return pd.DataFrame.from_records(self.dataRecords)

    @data.setter
    def data(self, val):
        self.dataRecords = val.to_dict(orient="records")

    def on_drag(self, callback):
        self.observe(callback, names=["fromValue", "toValue"])


@widgets.register
class TextArea(TextBaseWidget):
    _view_name = Unicode("TextAreaView").tag(sync=True)
    _model_name = Unicode("TextAreaModel").tag(sync=True)

    def on_text_changed(self, callback):
        self.observe(callback, names=["value"])


@widgets.register
class Text(TextBaseWidget):
    _view_name = Unicode("TextView").tag(sync=True)
    _model_name = Unicode("TextModel").tag(sync=True)



import mimetypes

import ipywidgets as widgets
from traitlets import Bool, Float, Int, TraitType, Unicode

from d3vis_ipynb.base_widget import BaseWidget


class _Media(BaseWidget):
    value = TraitType().tag(sync=True)
    format = Unicode().tag(sync=True)
    width = Int().tag(sync=True)
    height = Int().tag(sync=True)

    def __init__(self, file, type, **kwargs):
        filename = ""
        read_file = None

        if getattr(file, "read", None) is not None:
            read_file = file.read()
            filename = file.name
        else:
            with open(file, "rb") as f:
                read_file = f.read()
                filename = file
        self.format = self._guess_format(type, filename)
        self.value = read_file
        super().__init__(**kwargs)

    def _guess_format(self, tag, file):
        name = getattr(file, "name", None)
        name = name or file

        try:
            mtype, _ = mimetypes.guess_type(name)
            if not mtype.startswith("{}/".format(tag)):
                return ""

            return mtype[len("{}/".format(tag)) :]
        except Exception:
            return ""


@widgets.register
class Video(_Media):
    _view_name = Unicode("VideoView").tag(sync=True)
    _model_name = Unicode("VideoModel").tag(sync=True)

    _play = Bool().tag(sync=True)
    _pause = Bool().tag(sync=True)
    _duration = Float().tag(sync=True)
    _seekTo = Float().tag(sync=True)
    _seeked = Bool().tag(sync=True)

    _currentTime = Float().tag(sync=True)
    controls = Bool().tag(sync=True)
    loop = Bool().tag(sync=True)
    muted = Bool().tag(sync=True)
    volume = Float().tag(sync=True)

    def __init__(self, file, controls=True, loop=True, muted=False, volume=1, **kwargs):
        self.controls = controls
        self.loop = loop
        self.muted = muted
        self.volume = volume
        super().__init__(file, "video", **kwargs)

    def play(self):
        self._play = not self._play

    def pause(self):
        self._pause = not self._pause

    @property
    def duration(self):
        return self._duration

    def on_duration_set(self, callback):
        self.observe(callback, names=["_duration"])

    def on_current_time_change(self, callback):
        self.observe(callback, names=["_currentTime"])

    def seekTo(self, time):
        self._seekTo = time
        self._seeked = not self._seeked

    @property
    def currentTime(self):
        return self._currentTime


@widgets.register
class Image(_Media):
    _view_name = Unicode("ImageView").tag(sync=True)
    _model_name = Unicode("ImageModel").tag(sync=True)

    def __init__(self, file, **kwargs):
        super().__init__(file, "image", **kwargs)



import ipywidgets as widgets
from traitlets import Unicode
from ._version import NPM_PACKAGE_RANGE


class BaseWidget(widgets.DOMWidget):
    _view_module = Unicode("d3vis_ipynb").tag(sync=True)
    _model_module = Unicode("d3vis_ipynb").tag(sync=True)
    _view_module_version = Unicode(NPM_PACKAGE_RANGE).tag(sync=True)
    _model_module_version = Unicode(NPM_PACKAGE_RANGE).tag(sync=True)

    elementId = Unicode().tag(sync=True)
